# A3 solutions version
# Lillian Lee (LJL2)
# April 2017, minor updates to to make compliant with Python 3 in 2018

"""CS1110 2017 Spring assignment on processing and comparing text samples"""

import random
import a3given_2017sp_soln as a3given
import math  # for the log function


# This function is the core of a3given.sample_from_source_file
def make_sample(textstring, label):
    """Returns: a new Sample where all its attributes are filled in correctly
    with respect to the string textstring.

    Preconditions: textstring is a nonempty string of English text.  Words are
    delimited by whitespace, are all lowercase, and all punctuation has been
    removed."""
    # Because we haven't talked about writing your own methods yet,
    # this function directly modifies the attributes of a Sample that
    # is created. This is UNORTHODOX Python; later, we'll learn the "right"
    # way to do this, which involves writing the __init__ method for Sample.

    sample = a3given.Sample()
    sample.label = label
    sample.text = textstring.split()
    sample.length = len(sample.text)

    ## Using three separate passes through sample.text, for code simplicity.
    ## Could have done one pass, with the unigram, bigram and trigram data
    ## updated simultaneously, but the code would probably be hard to read.

    for token in sample.text:
        if token in sample.unigram_counts:
            # You can write "x += 1" as shorthand for "x = x + 1"
            sample.unigram_counts[token] += 1
        else:
            sample.unigram_counts[token] = 1

        ## Alternate solution:
        ## More compact code, but inefficient because each use of <count> makes
        ## its own pass through sample.text
        # sample.unigram_counts[token] = sample.text.count(token)

    # Compile the bigram dictionary
    for i in list(range(sample.length - 1)):   # could omit the "list()" here
        start_word = sample.text[i]
        next_word = sample.text[i+1]
        if start_word in sample.bigram_dict:
            sample.bigram_dict[start_word].append(next_word)
        else:
            sample.bigram_dict[start_word] = [next_word]  # new 1-item list

    # Compile the trigram dictionary
    for i in list(range(sample.length - 2)):   # could omit the "list()" here
        start_words = sample.text[i] + " " + sample.text[i+1]
        next_word = sample.text[i+2]

        if start_words in sample.trigram_dict:
            sample.trigram_dict[start_words].append(next_word)
        else:
            sample.trigram_dict[start_words] = [next_word]

    return sample


def merge(samplelist, label):
    """Returns a new Sample whose text is the concatenation of the text for
    all the Samples in samplelist and whose label is <label>.

    Precondition: samplelist is a list of Samples of length >= 2.
    """

    # Several equivalently valid approaches were possible.

    newtext = " ".join(samplelist[0].text)
    for i in list(range(1,len(samplelist))):   # could omit the "list()" here
        newtext += (" " + " ".join(samplelist[i].text))
    return make_sample(newtext, label)


def diffs(s1, s2, k):
    """Returns: [justs1, justs2, ranked], where
    justs1 is a list of words occurring at least k times in s1.text but never in s2.text
        (order doesn't matter);
    justs2 is a list of words occurring at least k times in s2.text but never in s1.text
        (order doesn't matter)
    ranked is a list of items [word, logprobratio] where word occurs in both s1.text
        and s2.text, and logratio is the score mentioned in the assignment
        handout, rounded to 3 decimal digits using round(f, 3) where f is a float
        The items in ranked are sorted by logprobratio, highest first.

    In "real life", justs1 would be a list of [w, frequency of w] and similarly
    for justs2, but requiring students to do so would have lengthened the
    assignment and not tested new concepts beyond what getting <ranked> right
    does.

    Precondition: s1 and s2 are Samples."""

    justs1 = []
    justs2 = []
    ranked = []

    for word in s1.unigram_counts:
        if word not in s2.unigram_counts:
            if s1.unigram_counts[word] >= k:
                justs1.append(word)
        else:
            # Students do not need to have try/except here; I was watching for
            # numerical issues.
            try:
                # Converting to float for backwards compatibility with Python 2
                p1 = float(s1.unigram_counts[word])/s1.length
                p2 = float(s2.unigram_counts[word])/s2.length
                logprobratio = round(math.log(p1/p2),3)
                ranked.append([word, logprobratio])
            except:
                # Something wrong happened; print what we were working on
                print("word ", word, " p1: ", p1, " p2 ", p2)

    for word in s2.unigram_counts:
        if word not in s1.unigram_counts:
            if s2.unigram_counts[word] >= k:
                justs2.append(word)

    a3given.sort_weighted_list(ranked)
    return [justs1, justs2, ranked]


def bigram_generation(s, k):
    """Returns <= k-token string according to the bigram dictionary of Sample s.
    Fewer than k tokens are generated if the generator chooses a token that
    is not followed by any token in s.text (in particular, the last token might
    have that property.)

    Preconditions: k>1 is an int."""

    output = random.choice(s.text)
    prev_word = output  # prev_word is always the most recently generated word

    for c in list(range(k-1)):   # could omit the "list()" here
        if prev_word in s.bigram_dict:
            new_word = random.choice(s.bigram_dict[prev_word])
            output += (" " + new_word)
            prev_word = new_word

    return output


def trigram_generation(s, k):
    """Generate <= k-token string according to the trigram dictionary of Sample
    s.

    Fewer than k tokens are generated if the generator chooses a token that
    is not followed by any token in s.text (in particular, the last token might
    have that property.)

    Preconditions: k>2 is an int, s is a Sample."""

    # Start with selection of location of starting bigram "w1 w2" from the text
    i = random.randint(0, s.length-3)  # randint *includes* second argument position.

    output = s.text[i] + " " + s.text[i+1]
    prev_word_list = [s.text[i], s.text[i+1]]  # list of two most recently generated

    for c in list(range(k-2)):  # could omit the "list()" here
        key = ' '.join(prev_word_list)
        if key in s.trigram_dict:
            new_word = random.choice(s.trigram_dict[key])
            output += (" " + new_word)

            # set up for next time around
            prev_word_list[0] = prev_word_list[1]
            prev_word_list[1] = new_word

    return output
